@@ -8,13 +8,24 @@ extern crate unreachable;
 
 use std::cmp::Ordering;
 use std::error::Error;
-use std::ops::{Add, AddAssign, Deref, DerefMut, Div, DivAssign, Mul, MulAssign, Neg, Rem, RemAssign, Sub, SubAssign};
+use std::ops::{Add, AddAssign, Deref, DerefMut, Div, DivAssign, Mul, MulAssign, Neg, Rem,
+               RemAssign, Sub, SubAssign};
 use std::hash::{Hash, Hasher};
 use std::fmt;
 use std::io;
+use std::mem;
 use unreachable::unreachable;
 use num_traits::Float;
 
+// masks for the parts of the IEEE 754 float
+const SIGN_MASK: u64 = 0x8000000000000000u64;
+const EXP_MASK: u64 = 0x7ff0000000000000u64;
+const MAN_MASK: u64 = 0x000fffffffffffffu64;
+
+// canonical raw bit patterns (for hashing)
+const CANONICAL_NAN_BITS: u64 = 0x7ff8000000000000u64;
+const CANONICAL_ZERO_BITS: u64 = 0x0u64;
+
 /// A wrapper around Floats providing an implementation of Ord and Hash.
 ///
 /// NaN is sorted as *greater* than all other values and *equal*
@@ -66,11 +77,7 @@ impl<T: Float + PartialOrd> Ord for OrderedFloat<T> {
 impl<T: Float + PartialEq> PartialEq for OrderedFloat<T> {
     fn eq(&self, other: &OrderedFloat<T>) -> bool {
         if self.as_ref().is_nan() {
-            if other.as_ref().is_nan() {
-                true
-            } else {
-                false
-            }
+            if other.as_ref().is_nan() { true } else { false }
         } else if other.as_ref().is_nan() {
             false
         } else {
@@ -128,7 +135,7 @@ impl<T: Float> DerefMut for OrderedFloat<T> {
     }
 }
 
-impl<T: Float + PartialEq> Eq for OrderedFloat<T> { }
+impl<T: Float + PartialEq> Eq for OrderedFloat<T> {}
 
 /// A wrapper around Floats providing an implementation of Ord and Hash.
 ///
@@ -204,7 +211,7 @@ impl Into<f64> for NotNaN<f64> {
 
 /// Creates a NotNaN value from a Float.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl<T: Float> From<T> for NotNaN<T> {
     fn from(v: T) -> Self {
         assert!(!v.is_nan());
@@ -232,35 +239,38 @@ impl<T: Float> Add for NotNaN<T> {
 
 /// Adds a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl<T: Float> Add<T> for NotNaN<T> {
     type Output = Self;
 
     fn add(self, other: T) -> Self {
         assert!(!other.is_nan());
-        NotNaN(self.0 + other)
+        NotNaN::new(self.0 + other).expect("Addition resulted in NaN")
     }
 }
 
 impl AddAssign for NotNaN<f64> {
     fn add_assign(&mut self, other: Self) {
         self.0 += other.0;
+        assert!(!self.0.is_nan(), "Addition resulted in NaN")
     }
 }
 
 impl AddAssign for NotNaN<f32> {
     fn add_assign(&mut self, other: Self) {
         self.0 += other.0;
+        assert!(!self.0.is_nan(), "Addition resulted in NaN")
     }
 }
 
 /// Adds a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl AddAssign<f64> for NotNaN<f64> {
     fn add_assign(&mut self, other: f64) {
         assert!(!other.is_nan());
         self.0 += other;
+        assert!(!self.0.is_nan(), "Addition resulted in NaN")
     }
 }
 
@@ -271,6 +281,7 @@ impl AddAssign<f32> for NotNaN<f32> {
     fn add_assign(&mut self, other: f32) {
         assert!(!other.is_nan());
         self.0 += other;
+        assert!(!self.0.is_nan(), "Addition resulted in NaN")
     }
 }
 
@@ -278,51 +289,55 @@ impl<T: Float> Sub for NotNaN<T> {
     type Output = Self;
 
     fn sub(self, other: Self) -> Self {
-        NotNaN(self.0 - other.0)
+        NotNaN::new(self.0 - other.0).expect("Subtraction resulted in NaN")
     }
 }
 
 /// Subtracts a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl<T: Float> Sub<T> for NotNaN<T> {
     type Output = Self;
 
     fn sub(self, other: T) -> Self {
         assert!(!other.is_nan());
-        NotNaN(self.0 - other)
+        NotNaN::new(self.0 - other).expect("Subtraction resulted in NaN")
     }
 }
 
 impl SubAssign for NotNaN<f64> {
     fn sub_assign(&mut self, other: Self) {
         self.0 -= other.0;
+        assert!(!self.0.is_nan(), "Subtraction resulted in NaN")
     }
 }
 
 impl SubAssign for NotNaN<f32> {
     fn sub_assign(&mut self, other: Self) {
         self.0 -= other.0;
+        assert!(!self.0.is_nan(), "Subtraction resulted in NaN")
     }
 }
 
 /// Subtracts a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl SubAssign<f64> for NotNaN<f64> {
     fn sub_assign(&mut self, other: f64) {
         assert!(!other.is_nan());
         self.0 -= other;
+        assert!(!self.0.is_nan(), "Subtraction resulted in NaN")
     }
 }
 
 /// Subtracts a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl SubAssign<f32> for NotNaN<f32> {
     fn sub_assign(&mut self, other: f32) {
         assert!(!other.is_nan());
         self.0 -= other;
+        assert!(!self.0.is_nan(), "Subtraction resulted in NaN")
     }
 }
 
@@ -330,31 +345,33 @@ impl<T: Float> Mul for NotNaN<T> {
     type Output = Self;
 
     fn mul(self, other: Self) -> Self {
-        NotNaN(self.0 * other.0)
+        NotNaN::new(self.0 * other.0).expect("Multiplication resulted in NaN")
     }
 }
 
 /// Multiplies a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl<T: Float> Mul<T> for NotNaN<T> {
     type Output = Self;
 
     fn mul(self, other: T) -> Self {
         assert!(!other.is_nan());
-        NotNaN(self.0 * other)
+        NotNaN::new(self.0 * other).expect("Multiplication resulted in NaN")
     }
 }
 
 impl MulAssign for NotNaN<f64> {
     fn mul_assign(&mut self, other: Self) {
         self.0 *= other.0;
+        assert!(!self.0.is_nan(), "Multiplication resulted in NaN")
     }
 }
 
 impl MulAssign for NotNaN<f32> {
     fn mul_assign(&mut self, other: Self) {
         self.0 *= other.0;
+        assert!(!self.0.is_nan(), "Multiplication resulted in NaN")
     }
 }
 
@@ -370,11 +387,12 @@ impl MulAssign<f64> for NotNaN<f64> {
 
 /// Multiplies a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl MulAssign<f32> for NotNaN<f32> {
     fn mul_assign(&mut self, other: f32) {
         assert!(!other.is_nan());
         self.0 *= other;
+        assert!(!self.0.is_nan(), "Multiplication resulted in NaN")
     }
 }
 
@@ -382,51 +400,55 @@ impl<T: Float> Div for NotNaN<T> {
     type Output = Self;
 
     fn div(self, other: Self) -> Self {
-        NotNaN(self.0 / other.0)
+        NotNaN::new(self.0 / other.0).expect("Division resulted in NaN")
     }
 }
 
 /// Divides a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl<T: Float> Div<T> for NotNaN<T> {
     type Output = Self;
 
     fn div(self, other: T) -> Self {
         assert!(!other.is_nan());
-        NotNaN(self.0 / other)
+        NotNaN::new(self.0 / other).expect("Division resulted in NaN")
     }
 }
 
 impl DivAssign for NotNaN<f64> {
     fn div_assign(&mut self, other: Self) {
         self.0 /= other.0;
+        assert!(!self.0.is_nan(), "Division resulted in NaN")
     }
 }
 
 impl DivAssign for NotNaN<f32> {
     fn div_assign(&mut self, other: Self) {
         self.0 /= other.0;
+        assert!(!self.0.is_nan(), "Division resulted in NaN")
     }
 }
 
 /// Divides a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl DivAssign<f64> for NotNaN<f64> {
     fn div_assign(&mut self, other: f64) {
         assert!(!other.is_nan());
         self.0 /= other;
+        assert!(!self.0.is_nan(), "Division resulted in NaN")
     }
 }
 
 /// Divides a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl DivAssign<f32> for NotNaN<f32> {
     fn div_assign(&mut self, other: f32) {
         assert!(!other.is_nan());
         self.0 /= other;
+        assert!(!self.0.is_nan(), "Division resulted in NaN")
     }
 }
 
@@ -434,51 +456,55 @@ impl<T: Float> Rem for NotNaN<T> {
     type Output = Self;
 
     fn rem(self, other: Self) -> Self {
-        NotNaN(self.0 % other.0)
+        NotNaN::new(self.0 % other.0).expect("Rem resulted in NaN")
     }
 }
 
 /// Calculates `%` with a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl<T: Float> Rem<T> for NotNaN<T> {
     type Output = Self;
 
     fn rem(self, other: T) -> Self {
         assert!(!other.is_nan());
-        NotNaN(self.0 % other)
+        NotNaN::new(self.0 % other).expect("Rem resulted in NaN")
     }
 }
 
 impl RemAssign for NotNaN<f64> {
     fn rem_assign(&mut self, other: Self) {
         self.0 %= other.0;
+        assert!(!self.0.is_nan(), "Rem resulted in NaN")
     }
 }
 
 impl RemAssign for NotNaN<f32> {
     fn rem_assign(&mut self, other: Self) {
         self.0 %= other.0;
+        assert!(!self.0.is_nan(), "Rem resulted in NaN")
     }
 }
 
 /// Calculates `%=` with a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl RemAssign<f64> for NotNaN<f64> {
     fn rem_assign(&mut self, other: f64) {
         assert!(!other.is_nan());
         self.0 %= other;
+        assert!(!self.0.is_nan(), "Rem resulted in NaN")
     }
 }
 
 /// Calculates `%=` with a float directly.
 ///
-/// Panics if the provided value is NaN.
+/// Panics if the provided value is NaN or the computation results in NaN
 impl RemAssign<f32> for NotNaN<f32> {
     fn rem_assign(&mut self, other: f32) {
         assert!(!other.is_nan());
         self.0 %= other;
+        assert!(!self.0.is_nan(), "Rem resulted in NaN")
     }
 }
 
@@ -486,7 +512,7 @@ impl<T: Float> Neg for NotNaN<T> {
     type Output = Self;
 
     fn neg(self) -> Self {
-        NotNaN(-self.0)
+        NotNaN::new(-self.0).expect("Negation resulted in NaN")
     }
 }
 
@@ -496,7 +522,7 @@ pub struct FloatIsNaN;
 
 impl Error for FloatIsNaN {
     fn description(&self) -> &str {
-        return "NotNaN constructed with NaN"
+        return "NotNaN constructed with NaN";
     }
 }
 
@@ -514,78 +540,82 @@ impl Into<io::Error> for FloatIsNaN {
 
 #[inline]
 fn hash_float<F: Float, H: Hasher>(f: &F, state: &mut H) {
-    let (man, exp, sign) = f.integer_decode();
-    if man == 0 {
-        // Consolidate the representation of zero, whether signed or not
-        // The IEEE standard considers positive and negative zero to be equal
-        0
-    } else {
-        (man ^ ((exp as u64) << 48) ^ sign as u64)
-    }.hash(state)
+    raw_double_bits(f).hash(state);
 }
 
-#[cfg(feature = "rustc-serialize")]
-mod impl_rustc {
-    extern crate rustc_serialize;
-    use self::rustc_serialize::{Encodable, Encoder, Decodable, Decoder};
-    use super::{OrderedFloat, NotNaN};
-    use std::error::Error;
-    use num_traits::Float;
-
-    impl<T: Float + Encodable> Encodable for OrderedFloat<T> {
-        fn encode<S: Encoder>(&self, s: &mut S) -> Result<(), S::Error> {
-            self.0.encode(s)
-        }
+#[inline]
+fn raw_double_bits<F: Float>(f: &F) -> u64 {
+    if f.is_nan() {
+        return CANONICAL_NAN_BITS;
     }
 
-    impl<T: Float + Decodable> Decodable for OrderedFloat<T> {
-        fn decode<D: Decoder>(d: &mut D) -> Result<Self, D::Error> {
-            T::decode(d).map(OrderedFloat)
-        }
+    let (man, exp, sign) = f.integer_decode();
+    if man == 0 {
+        return CANONICAL_ZERO_BITS;
     }
 
-    impl<T: Float + Encodable> Encodable for NotNaN<T> {
-        fn encode<S: Encoder>(&self, s: &mut S) -> Result<(), S::Error> {
-            self.0.encode(s)
-        }
-    }
-
-    impl<T: Float + Decodable> Decodable for NotNaN<T> {
-        fn decode<D: Decoder>(d: &mut D) -> Result<Self, D::Error> {
-            T::decode(d).and_then(|v| NotNaN::new(v).map_err(|e| d.error(e.description())))
-        }
-    }
+    let exp_u64 = unsafe { mem::transmute::<i16, u16>(exp) } as u64;
+    let sign_u64 = if sign > 0 { 1u64 } else { 0u64 };
+    (man & MAN_MASK) | ((exp_u64 << 52) & EXP_MASK) | ((sign_u64 << 63) & SIGN_MASK)
 }
 
 #[cfg(feature = "serde")]
 mod impl_serde {
     extern crate serde;
     use self::serde::{Serialize, Serializer, Deserialize, Deserializer};
-    use self::serde::de::Error;
+    use self::serde::de::{Error, Unexpected};
     use super::{OrderedFloat, NotNaN};
     use num_traits::Float;
+    use std::f64;
+
+    #[cfg(test)]
+    extern crate serde_test;
+    #[cfg(test)]
+    use self::serde_test::{Token, assert_tokens, assert_de_tokens_error};
 
     impl<T: Float + Serialize> Serialize for OrderedFloat<T> {
-        fn serialize<S: Serializer>(&self, s: &mut S) -> Result<(), S::Error> {
+        fn serialize<S: Serializer>(&self, s: S) -> Result<S::Ok, S::Error> {
             self.0.serialize(s)
         }
     }
 
-    impl<T: Float + Deserialize> Deserialize for OrderedFloat<T> {
-        fn deserialize<D: Deserializer>(d: &mut D) -> Result<Self, D::Error> {
+    impl<'de, T: Float + Deserialize<'de>> Deserialize<'de> for OrderedFloat<T> {
+        fn deserialize<D: Deserializer<'de>>(d: D) -> Result<Self, D::Error> {
             T::deserialize(d).map(OrderedFloat)
         }
     }
 
     impl<T: Float + Serialize> Serialize for NotNaN<T> {
-        fn serialize<S: Serializer>(&self, s: &mut S) -> Result<(), S::Error> {
+        fn serialize<S: Serializer>(&self, s: S) -> Result<S::Ok, S::Error> {
             self.0.serialize(s)
         }
     }
 
-    impl<T: Float + Deserialize> Deserialize for NotNaN<T> {
-        fn deserialize<D: Deserializer>(d: &mut D) -> Result<Self, D::Error> {
-            T::deserialize(d).and_then(|v| NotNaN::new(v).map_err(|_| <D::Error as Error>::invalid_value("value cannot be NaN")))
+    impl<'de, T: Float + Deserialize<'de>> Deserialize<'de> for NotNaN<T> {
+        fn deserialize<D: Deserializer<'de>>(d: D) -> Result<Self, D::Error> {
+            let float = T::deserialize(d)?;
+            NotNaN::new(float).map_err(|_| {
+                Error::invalid_value(Unexpected::Float(f64::NAN), &"float (but not NaN)")
+            })
         }
     }
+
+    #[test]
+    fn test_ordered_float() {
+        let float = OrderedFloat(1.0f64);
+        assert_tokens(&float, &[Token::F64(1.0)]);
+    }
+
+    #[test]
+    fn test_not_nan() {
+        let float = NotNaN(1.0f64);
+        assert_tokens(&float, &[Token::F64(1.0)]);
+    }
+
+    #[test]
+    fn test_fail_on_nan() {
+        assert_de_tokens_error::<NotNaN<f64>>(
+            &[Token::F64(f64::NAN)],
+            "invalid value: floating point `NaN`, expected float (but not NaN)");
+    }
 }
